# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


from ..keras.models import Model as KerasModel
from ..keras.layers import Input, concatenate, Lambda
from ..utils import is_keras_standalone
from tensorflow import image
import tensorflow as tf

from ..model.layers import layer_from_config
from ..utils import deep_copy


class BaseModelLoader:
    """
    The base class of all model loader. A model loader will create a
    neuralnetwork model with a predefined architecture.
    For example, UnetModelLoader creates a neural network with Unet structure.

    Raises
    ------
    NotImplementedError
        `load` method needs to be implemented in
        children classes
    """
    def load():
        raise NotImplementedError()

    def __init__(self, architecture, input_params):
        """
        Initialize a model loader

        Parameters
        ----------
        architecture : dict
            Configuration for the network architecture

            Example:
            ```
            {
                "type": "Sequential",
                "layers": [
                    {
                    "class_name": "Conv2D",
                    "config": {
                        "filters": 4,
                        "kernel_size": 3,
                        "activation": "relu",
                        "padding": "same"
                    },
                    {
                    "class_name": "Dense",
                    "config": {
                        "units": 2,
                        "activation": "sigmoid"
                    },
                ]
            }
            ```

        input_params : dict
            Configuration for the input layer

            Example
            ```
            {
                "shape": [128, 128, 3]
            }
            ```


        """
        if 'layers' in architecture:
            self._layers = deep_copy(architecture['layers'])

        self._input_params = deep_copy(input_params)


class SequentialModelLoader(BaseModelLoader):
    """
    Create a sequential network from list of layers
    """

    def load(self):
        """

        Returns
        -------
        tensorflow.keras.models.Model
            A neural network of sequential layers
            from the configured layer list.
        """
        layers = [Input(**self._input_params)]

        for i, layer in enumerate(self._layers):
            next_tensor = layer_from_config(layer)
            layers.append(next_tensor(layers[i]))
        return KerasModel(inputs=layers[0], outputs=layers[-1])


class UnetModelLoader(BaseModelLoader):
    """
    Create a unet neural network from layers

    Raises
    ------
    NotImplementedError
        volumn adjustment in skip connection
        doesn't support for 3D unet
    """

    def load(self):
        """
        Load the unet neural network.

        Example of Configuration for `layers`:
        ```
        [
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 4,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "name": "conv_1",
                "class_name": "Conv2D",
                "config": {
                    "filters": 4,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "class_name": "MaxPooling2D"
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 8,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "name": "conv_2",
                "class_name": "Conv2D",
                "config": {
                    "filters": 8,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            ...
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 64,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "name": "conv_5",
                "class_name": "Conv2D",
                "config": {
                    "filters": 64,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "class_name": "MaxPooling2D"
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 128,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 128,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "name": "conv_T_1",
                "class_name": "Conv2DTranspose",
                "config": {
                    "filters": 32,
                    "kernel_size": 3,
                    "strides": [
                        2,
                        2
                    ],
                    "padding": "same"
                }
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 64,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                },
                "inputs": [
                    "conv_T_1",
                    "conv_5"
                ]
            },
            ...
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 16,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "name": "conv_T_4",
                "class_name": "Conv2DTranspose",
                "config": {
                    "filters": 4,
                    "kernel_size": 3,
                    "strides": [
                        2,
                        2
                    ],
                    "padding": "same"
                }
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 8,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                },
                "inputs": [
                    "conv_T_4",
                    "conv_2"
                ]
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 8,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "name": "conv_T_5",
                "class_name": "Conv2DTranspose",
                "config": {
                    "filters": 2,
                    "kernel_size": 3,
                    "strides": [
                        2,
                        2
                    ],
                    "padding": "same"
                }
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 4,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                },
                "inputs": [
                    "conv_T_5",
                    "conv_1"
                ]
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 4,
                    "kernel_size": 3,
                    "activation": "relu",
                    "padding": "same"
                }
            },
            {
                "class_name": "Conv2D",
                "config": {
                    "filters": 1,
                    "kernel_size": 1,
                    "activation": "sigmoid"
                }
            }
        ]
        ```

        Returns
        -------
        tensorflow.keras.models.Model
            A neural network with unet structure

        Raises
        ------
        NotImplementedError
            Doesn't support volumn adjustment in skip connection
        """
        layers = [Input(**self._input_params)]
        saved_input = {}

        for i, layer in enumerate(self._layers):
            next_tensor = layer_from_config(layer)

            if 'inputs' in layer:
                inputs = []
                size_factors = None
                for input_name in layer['inputs']:
                    if is_keras_standalone():  # pragma: no cover
                        # keras issue: convtranspose layer output shape are
                        # (None, None, None, filters)
                        if saved_input[
                                input_name].get_shape() != saved_input[
                                    input_name]._keras_shape:
                            saved_input[input_name].set_shape(
                                saved_input[input_name]._keras_shape)

                    if size_factors:
                        if size_factors == saved_input[
                                input_name].get_shape().as_list()[1:-1]:
                            next_input = saved_input[input_name]
                        else:
                            if len(size_factors) == 2:
                                if is_keras_standalone():  # pragma: no cover
                                    # Create the resize function
                                    def resize_tensor(
                                            input_tensor, resize_fn, new_size):
                                        return resize_fn(
                                            input_tensor,
                                            new_size,
                                            method='bilinear')

                                    # Put it in the lambda layer
                                    next_input = Lambda(
                                        resize_tensor,
                                        arguments={
                                            "resize_fn": image.resize,
                                            "new_size": size_factors}
                                    )(saved_input[input_name])
                                else:
                                    next_input = image.resize(
                                        saved_input[input_name],
                                        size_factors,
                                        method='bilinear')
                            else:
                                raise NotImplementedError(
                                    "Resize 3D tensor not implemented")
                        inputs.append(next_input)

                    else:
                        inputs.append(saved_input[input_name])
                        size_factors = saved_input[
                            input_name].get_shape().as_list()[1:-1]
                connected_input = concatenate(inputs)
            else:
                connected_input = layers[i]

            next_layer = next_tensor(connected_input)

            if 'normalizer' in layer:
                next_layer = layer_from_config(layer['normalizer'])(next_layer)

            if 'name' in layer:
                saved_input[layer['name']] = next_layer

            layers.append(next_layer)

        return KerasModel(inputs=layers[0], outputs=layers[-1])


# TODO: refactor this for Afreen
class Vnet(BaseModelLoader):
    """
    EXPERIMENTAL
    author: Afreen Mirza
    email: afreen.mirza@nmbu.no

    Create a vnet neural network from layers
    """

    def resize_by_axis(self, img, dim_1, dim_2, ax):
        resized_list = []
        # print(img.shape, ax, dim_1, dim_2)
        unstack_img_depth_list = tf.unstack(img, axis=ax)
        for j in unstack_img_depth_list:
            resized_list.append(
                image.resize(j, [dim_1, dim_2], method='bicubic'))
        stack_img = tf.stack(resized_list, axis=ax)
        # print(stack_img.shape)
        return stack_img

    def resize_along_dim(self, img, new_dim):
        dim_1, dim_2, dim_3 = new_dim

        resized_along_depth = self.resize_by_axis(img, dim_1, dim_2, 3)
        resized_along_width = self.resize_by_axis(
            resized_along_depth, dim_1, dim_3, 2)
        return resized_along_width

    def load(self):
        """
        Load the unet neural network. Use Conv3d

        Returns
        -------
        tensorflow.keras.models.Model
            A neural network with unet structure

        Raises
        ------
        NotImplementedError
            Does not support video and time-series image inputs
        """
        global next_input
        layers = [Input(**self._input_params)]
        saved_input = {}

        for i, layer in enumerate(self._layers):
            next_tensor = layer_from_config(layer)

            if 'inputs' in layer:
                inputs = []
                size_factors = None
                for input_name in layer['inputs']:
                    if size_factors:
                        if size_factors == saved_input[
                                input_name].get_shape().as_list()[1:-1]:
                            next_input = saved_input[input_name]
                        else:
                            if len(size_factors) == 2:
                                next_input = image.resize(
                                    saved_input[input_name],
                                    size_factors,
                                    # preserve_aspect_ratio=True,
                                    method='bilinear')
                            elif len(size_factors) == 3:

                                next_input = self.resize_along_dim(
                                    saved_input[input_name],
                                    size_factors
                                )

                            else:
                                raise NotImplementedError(
                                    "Image shape is not supported ")
                        inputs.append(next_input)

                    else:
                        inputs.append(saved_input[input_name])
                        size_factors = saved_input[
                            input_name].get_shape().as_list()[1:-1]
                connected_input = concatenate(inputs)
            else:
                connected_input = layers[i]

            next_layer = next_tensor(connected_input)

            if 'normalizer' in layer:
                next_layer = layer_from_config(layer['normalizer'])(next_layer)

            if 'name' in layer:
                saved_input[layer['name']] = next_layer

            layers.append(next_layer)

        return KerasModel(inputs=layers[0], outputs=layers[-1])


class DenseModelLoader(BaseModelLoader):
    def load():
        pass


class ModelLoaderFactory:
    _loaders = {
        'Sequential': SequentialModelLoader,
        'Unet': UnetModelLoader,
        'Vnet': Vnet,
        'Dense': DenseModelLoader
    }

    @classmethod
    def create(cls, architecture, input_params):
        model_type = architecture['type']

        if model_type in cls._loaders:
            return cls._loaders[model_type](architecture, input_params)
        else:
            raise ValueError('Invalid Architecture {}'.format(model_type))

    @classmethod
    def register(cls, model_type, loader):
        if not issubclass(loader, BaseModelLoader):
            raise ValueError(
                "The new loader has to be a subclass"
                + " of deoxys.loader.BaseModelLoader"
            )

        if model_type in cls._loaders:
            raise KeyError(
                "Duplicated key, please use another key name for this loader"
            )
        else:
            cls._loaders[model_type] = loader


def load_architecture(architecture, input_params):
    loader = ModelLoaderFactory.create(architecture, input_params)
    return loader.load()


def register_architecture(model_type, loader):
    ModelLoaderFactory.register(model_type, loader)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
import numpy as np
from deoxys.data import Preprocessors, BasePreprocessor, \
    SingleChannelPreprocessor, WindowingPreprocessor, \
    KerasImagePreprocessorX, KerasImagePreprocessorY, \
    HounsfieldWindowingPreprocessor, ImageNormalizerPreprocessor, \
    UnetPaddingPreprocessor, ChannelRemoval, ChannelSelector, \
    ImageAffineTransformPreprocessor, ImageAugmentation2D, \
    ImageAugmentation3D, \
    preprocessor_from_config
from deoxys.customize import register_preprocessor, \
    unregister_preprocessor, custom_preprocessor
from deoxys.utils import Singleton


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture
def preprocessor_class():
    class TestPreprocessor(BasePreprocessor):
        pass

    yield TestPreprocessor


def test_is_singleton():
    preprocessors_instance = Preprocessors()
    another_instance = Preprocessors()

    assert preprocessors_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_preprocessor("TestPreprocessor", object)


def test_register_preprocessor_success(preprocessor_class):
    register_preprocessor("TestPreprocessor", preprocessor_class)

    assert Preprocessors(
    )._preprocessors["TestPreprocessor"] is preprocessor_class


def test_register_duplicate_preprocessor(preprocessor_class):
    register_preprocessor("TestPreprocessor", preprocessor_class)

    with pytest.raises(KeyError):
        register_preprocessor("TestPreprocessor", preprocessor_class)


def test_unregister_preprocessor(preprocessor_class):
    register_preprocessor("TestPreprocessor", preprocessor_class)

    assert Preprocessors(
    )._preprocessors["TestPreprocessor"] is preprocessor_class
    unregister_preprocessor("TestPreprocessor")

    assert "TestPreprocessor" not in Preprocessors()._preprocessors


def test_decorator():
    @custom_preprocessor
    class TestPreprocessor2(BasePreprocessor):
        pass

    assert Preprocessors(
    ).preprocessors["TestPreprocessor2"] is TestPreprocessor2


def test_preprocessor_from_config():
    pp = preprocessor_from_config({
        'class_name': 'SingleChannelPreprocessor'
    })

    assert callable(pp.transform)


def test_single_channel_preprocessor():
    scp = SingleChannelPreprocessor()

    x, y = np.arange(100).reshape(4, 5, 5), np.arange(100).reshape(4, 5, 5)
    expected_x = np.arange(100).reshape(4, 5, 5, 1)
    expected_y = np.arange(100).reshape(4, 5, 5, 1)

    output_x, output_y = scp.transform(x, y)
    assert x.shape == (4, 5, 5), 'X remains unchanged'
    assert y.shape == (4, 5, 5), 'Y remains unchanged'
    assert expected_x.shape == output_x.shape
    assert expected_y.shape == output_y.shape
    assert np.all(expected_x == output_x)
    assert np.all(expected_y == output_y)


def test_windowing_preprocessor():
    wp = WindowingPreprocessor(50, 80, 0)

    y = np.arange(200).reshape(4, 5, 5, 2)
    x_first_channel = np.arange(100).reshape(4, 5, 5)
    x_second_channel = np.arange(100).reshape(4, 5, 5)
    x = np.stack([x_first_channel, x_second_channel], axis=-1)

    expected_y = np.arange(200).reshape(4, 5, 5, 2)
    expected_x_first_channel = np.concatenate(
        ([-40] * 10,
         np.arange(-40, 40),
         [40] * 10)
    ).reshape(4, 5, 5)
    expected_x_second_channel = np.arange(100).reshape(4, 5, 5)
    expected_x = np.stack(
        [expected_x_first_channel, expected_x_second_channel], axis=-1)

    ouput_x, ouput_y = wp.transform(x, y)

    assert ouput_x.shape == expected_x.shape
    assert ouput_y.shape == expected_y.shape
    assert np.all(ouput_x == expected_x)
    assert np.all(ouput_y == expected_y)


def test_hounsfield_windowing_preprocessor():
    wp = HounsfieldWindowingPreprocessor(30, 80, 0, 20)

    y = np.arange(200).reshape(4, 5, 5, 2)
    x_first_channel = np.arange(100).reshape(4, 5, 5)
    x_second_channel = np.arange(100).reshape(4, 5, 5)
    x = np.stack([x_first_channel, x_second_channel], axis=-1)

    expected_y = np.arange(200).reshape(4, 5, 5, 2)
    expected_x_first_channel = np.concatenate(
        ([-40] * 10,
         np.arange(-40, 40),
         [40] * 10)
    ).reshape(4, 5, 5)
    expected_x_second_channel = np.arange(100).reshape(4, 5, 5)
    expected_x = np.stack(
        [expected_x_first_channel, expected_x_second_channel], axis=-1)

    ouput_x, ouput_y = wp.transform(x, y)

    assert ouput_x.shape == expected_x.shape
    assert ouput_y.shape == expected_y.shape
    assert np.all(ouput_x == expected_x)
    assert np.all(ouput_y == expected_y)


def test_image_normalizer_preprocessor():
    base_data = np.array([np.arange(30) for _ in range(5)])
    targets = np.array([np.arange(30) for _ in range(5)])

    normalized_image = np.zeros(30)
    normalized_image[20:] = 1
    normalized_image[10:20] = np.arange(10) / 10

    normalize_data = np.array([normalized_image for _ in range(5)])

    images = np.zeros((5, 5, 6, 2))
    images[..., 0] = base_data.reshape(5, 5, 6)
    images[..., 1] = base_data.reshape(5, 5, 6)

    expected_x = np.zeros((5, 5, 6, 2))
    expected_x[..., 0] = normalize_data.reshape(5, 5, 6)
    expected_x[..., 1] = normalize_data.reshape(5, 5, 6)

    expected_y = np.array([np.arange(30) for _ in range(5)])

    output_x, output_y = ImageNormalizerPreprocessor(
        vmin=10, vmax=20).transform(images, targets)

    assert output_x.shape == expected_x.shape
    assert output_y.shape == expected_y.shape
    assert np.all(output_x == expected_x)
    assert np.all(output_y == expected_y)

    output_x, output_y = ImageNormalizerPreprocessor().transform(
        images, targets)

    assert output_x.shape == expected_x.shape
    assert output_y.shape == expected_y.shape
    assert np.allclose(output_x, images/29)
    assert np.all(output_y == expected_y)


def test_image_normalizer_preprocessor_per_channel():
    base_data = np.array([np.arange(30) for _ in range(5)])
    targets = np.array([np.arange(30) for _ in range(5)])

    normalized_image = np.zeros(30)
    normalized_image[20:] = 1
    normalized_image[10:20] = np.arange(10) / 10

    normalize_data = np.array([normalized_image for _ in range(5)])

    images = np.zeros((5, 5, 6, 3))
    images[..., 0] = base_data.reshape(5, 5, 6)
    images[..., 1] = base_data.reshape(5, 5, 6)
    images[..., 2] = base_data.reshape(5, 5, 6)

    expected_x = np.zeros((5, 5, 6, 3))
    expected_x[..., 0] = normalize_data.reshape(5, 5, 6)
    expected_x[..., 1] = normalize_data.reshape(5, 5, 6)
    expected_x[..., 2] = base_data.reshape(5, 5, 6) / 29

    expected_y = np.array([np.arange(30) for _ in range(5)])

    output_x, output_y = ImageNormalizerPreprocessor(
        [10, 10], [20, 20]).transform(images, targets)

    assert output_x.shape == expected_x.shape
    assert output_y.shape == expected_y.shape
    assert np.allclose(output_x, expected_x)
    assert np.all(output_y == expected_y)

    output_x, output_y = ImageNormalizerPreprocessor(
        [10, 20], [20, 30]).transform(images, targets)

    normalized_image = np.zeros(30)
    normalized_image[20:] = np.arange(10) / 10
    normalize_data = np.array([normalized_image for _ in range(5)])
    expected_x[..., 1] = normalize_data.reshape(5, 5, 6)

    assert output_x.shape == expected_x.shape
    assert output_y.shape == expected_y.shape
    assert np.allclose(output_x, expected_x)
    assert np.all(output_y == expected_y)


def test_padding_preprocessor():
    images = np.random.random((30, 40, 40, 2))
    targets = np.random.random((30, 40, 40, 1))

    output_x, output_y = UnetPaddingPreprocessor(
        depth=4).transform(images, targets)

    assert output_x.shape == (30, 48, 48, 2)
    assert output_y.shape == (30, 48, 48, 1)
    assert np.all(output_x[:, 4:44, 4:44] == images)
    assert np.all(output_y[:, 4:44, 4:44] == targets)

    images = np.random.random((30, 40, 40, 40, 2))
    targets = np.random.random((30, 40, 40, 40, 1))

    output_x, output_y = UnetPaddingPreprocessor(
        depth=4).transform(images, targets)

    assert output_x.shape == (30, 48, 48, 48, 2)
    assert output_y.shape == (30, 48, 48, 48, 1)
    assert np.all(output_x[:, 4:44, 4:44, 4:44] == images)
    assert np.all(output_y[:, 4:44, 4:44, 4:44] == targets)


def test_padding_preprocessor_unchanged():
    images = np.random.random((30, 48, 40, 2))
    targets = np.random.random((30, 48, 40, 1))

    output_x, output_y = UnetPaddingPreprocessor(
        depth=4).transform(images, targets)

    assert output_x.shape == (30, 48, 48, 2)
    assert output_y.shape == (30, 48, 48, 1)
    assert np.all(output_x[:, :, 4:44] == images)
    assert np.all(output_y[:, :, 4:44] == targets)

    images = np.random.random((30, 40, 48, 2))
    targets = np.random.random((30, 40, 48, 1))

    output_x, output_y = UnetPaddingPreprocessor(
        depth=4).transform(images, targets)

    assert output_x.shape == (30, 48, 48, 2)
    assert output_y.shape == (30, 48, 48, 1)
    assert np.all(output_x[:, 4:44] == images)
    assert np.all(output_y[:, 4:44] == targets)

    images = np.random.random((30, 40, 48, 48, 2))
    targets = np.random.random((30, 40, 48, 48, 1))

    output_x, output_y = UnetPaddingPreprocessor(
        depth=4).transform(images, targets)

    assert output_x.shape == (30, 48, 48, 48, 2)
    assert output_y.shape == (30, 48, 48, 48, 1)
    assert np.all(output_x[:, 4:44] == images)
    assert np.all(output_y[:, 4:44] == targets)

    images = np.random.random((30, 48, 40, 48, 2))
    targets = np.random.random((30, 48, 40, 48, 1))

    output_x, output_y = UnetPaddingPreprocessor(
        depth=4).transform(images, targets)

    assert output_x.shape == (30, 48, 48, 48, 2)
    assert output_y.shape == (30, 48, 48, 48, 1)
    assert np.all(output_x[:, :, 4:44] == images)
    assert np.all(output_y[:, :, 4:44] == targets)


def test_channel_removal():
    images = np.random.random((30, 40, 40, 3))
    targets = np.random.random((30, 40, 40, 1))

    output_x, output_y = ChannelRemoval(channel=1).transform(images, targets)
    assert output_x.shape == (30, 40, 40, 2)
    assert output_y.shape == (30, 40, 40, 1)
    assert np.all(output_x == images[..., [0, 2]])
    assert np.all(output_y == targets)

    images = np.random.random((30, 40, 40, 40, 3))
    targets = np.random.random((30, 40, 40, 40, 1))

    output_x, output_y = ChannelRemoval(
        channel=[1, 2]).transform(images, targets)
    assert output_x.shape == (30, 40, 40, 40, 1)
    assert output_y.shape == (30, 40, 40, 40, 1)
    assert np.all(output_x == images[..., [0]])
    assert np.all(output_y == targets)


def test_channel_selector():
    images = np.random.random((30, 40, 40, 3))
    targets = np.random.random((30, 40, 40, 1))

    output_x, output_y = ChannelSelector(
        channel=[0, 2]).transform(images, targets)
    assert output_x.shape == (30, 40, 40, 2)
    assert output_y.shape == (30, 40, 40, 1)
    assert np.all(output_x == images[..., [0, 2]])
    assert np.all(output_y == targets)

    images = np.random.random((30, 40, 40, 40, 3))
    targets = np.random.random((30, 40, 40, 40, 1))

    output_x, output_y = ChannelSelector(
        channel=0).transform(images, targets)
    assert output_x.shape == (30, 40, 40, 40, 1)
    assert output_y.shape == (30, 40, 40, 40, 1)
    assert np.all(output_x == images[..., [0]])
    assert np.all(output_y == targets)


def test_affine_transform():
    images = np.random.random((30, 40, 40, 3))
    targets = np.rint(np.random.random((30, 40, 40, 1)))

    output_x, output_y = ImageAffineTransformPreprocessor(
        flip_axis=1, shift=(0, 10)
    ).transform(images, targets)

    expected_x = np.zeros((30, 40, 40, 3))
    expected_x[:, :, :-10] = images[:, :, 10:]
    expected_x = np.flip(expected_x, axis=2)

    for i in range(30):
        expected_x[i] = expected_x[i].clip(
            images[i].min(axis=(0, 1)), images[i].max(axis=(0, 1)))

    expected_y = np.zeros((30, 40, 40, 1))
    expected_y[:, :, :-10] = targets[:, :, 10:]
    expected_y = np.flip(expected_y, axis=2)

    assert np.allclose(expected_x, output_x)
    assert np.all(expected_y == output_y)


def test_image_augmentation():
    images = np.random.random((30, 40, 45, 3))
    targets = np.rint(np.random.random((30, 40, 45, 1)))

    output_x, output_y = ImageAugmentation2D(rotation_range=14,
                                             zoom_range=(0.8, 1.2),
                                             flip_axis=1, shift_range=(0, 10)
                                             ).transform(images, targets)

    assert output_x.shape == (30, 40, 45, 3)
    assert output_y.shape == (30, 40, 45, 1)

    images = np.random.random((30, 40, 45, 50, 3))
    targets = np.rint(np.random.random((30, 40, 45, 50, 1)))

    output_x, output_y = ImageAugmentation3D(rotation_range=14,
                                             rotation_axis=0,
                                             zoom_range=(0.8, 1.2),
                                             flip_axis=1,
                                             shift_range=(0, 10, 5)
                                             ).transform(images, targets)

    assert output_x.shape == (30, 40, 45, 50, 3)
    assert output_y.shape == (30, 40, 45, 50, 1)


def test_keras_image_preprocessor():
    kpx = KerasImagePreprocessorX(scale_down=2, shuffle=False)
    kpy = KerasImagePreprocessorY(scale_down=4, shuffle=False)

    x, y = np.arange(100).reshape(
        4, 5, 5, 1), np.arange(100).reshape(4, 5, 5, 1)

    expected_x = np.arange(100).reshape(4, 5, 5, 1) / 2
    expected_y = np.arange(100).reshape(4, 5, 5, 1) / 4

    output_x, output_y = kpx.transform(x, y)
    output_x, output_y = kpy.transform(output_x, output_y)

    assert expected_x.shape == output_x.shape
    assert expected_y.shape == output_y.shape
    assert np.allclose(expected_x, output_x)
    assert np.all(expected_y == output_y)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
from deoxys.keras.optimizers import Adam
from deoxys.loaders import load_params
from deoxys.model.losses import BinaryFbetaLoss
from deoxys.model.metrics import BinaryFbeta
from deoxys.utils import read_file, load_json_config


def test_load_params():
    params = load_params(load_json_config(
        read_file('tests/json/model_param_config.json')))

    assert isinstance(params['optimizer'], Adam)
    assert isinstance(params['loss'], BinaryFbetaLoss)
    assert isinstance(params['metrics'][0], BinaryFbeta)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
import h5py
import os
import numpy as np
from deoxys.loaders import load_data
from deoxys.customize import custom_preprocessor
from deoxys.data import BasePreprocessor, HDF5Reader, H5Reader
from deoxys.utils import read_file, load_json_config, Singleton


def setup_data():
    # Setup the hdf5 files
    data = np.arange(25000)
    data = np.reshape(data, (1000, 5, 5))
    target = [1, 2, 3, 4, 5] * 200

    return data, target


def teardown(filename=''):
    if os.path.isfile(filename):
        os.remove(filename)


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture(scope='module', autouse=True)
def h5file():
    # Create standardized h5 file with 10 folds, named fold_x
    filename = 'data.h5'
    data, target = setup_data()

    hf = h5py.File(filename, 'w')

    for i in range(10):
        start, end = i * 100, (i + 1) * 100
        group = hf.create_group('fold_{0}'.format(i))
        group.create_dataset('input', data=data[start: end])
        group.create_dataset('target', data=target[start: end])

    hf.close()
    yield filename
    teardown(filename)


def check_equal_data_generator(actual, expected):
    assert actual.total_batch == expected.total_batch

    total_batch = actual.total_batch

    for i, (actual_batch, expected_batch) in enumerate(
            zip(actual.generate(), expected.generate())):
        if i >= total_batch:
            break

        assert np.all(actual_batch[0] == expected_batch[0])
        assert np.all(actual_batch[1] == expected_batch[1])


def test_load_data(h5file):
    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(self, x, y):
            return x + 1, y + 1

    actual_dr = load_data(load_json_config(
        read_file('tests/json/dataset_config.json')))

    expected_dr = HDF5Reader(
        filename=h5file,
        batch_size=8,
        preprocessors=PlusOnePreprocessor(),
        x_name='input',
        y_name='target',
        train_folds=[0, 1, 2],
        val_folds=[3],
        test_folds=[4, 5]
    )

    assert isinstance(actual_dr, HDF5Reader)
    assert isinstance(actual_dr.preprocessors[0], PlusOnePreprocessor)

    actual_train_data = actual_dr.train_generator
    actual_val_data = actual_dr.val_generator
    actual_test_data = actual_dr.test_generator

    expected_train_data = expected_dr.train_generator
    expected_val_data = expected_dr.val_generator
    expected_test_data = expected_dr.test_generator

    check_equal_data_generator(actual_train_data, expected_train_data)
    check_equal_data_generator(actual_val_data, expected_val_data)
    check_equal_data_generator(actual_test_data, expected_test_data)


def test_load_data_multi_preprocessor(h5file):
    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(self, x, y):
            return x + 1, y + 1

    actual_dr = load_data(load_json_config(
        read_file('tests/json/dataset_config_multi_preprocessor.json')))

    expected_dr = HDF5Reader(
        filename=h5file,
        batch_size=8,
        preprocessors=[PlusOnePreprocessor(), PlusOnePreprocessor()],
        x_name='input',
        y_name='target',
        train_folds=[0, 1, 2],
        val_folds=[3],
        test_folds=[4, 5]
    )

    assert isinstance(actual_dr, HDF5Reader)
    assert isinstance(actual_dr.preprocessors[0], PlusOnePreprocessor)
    assert isinstance(actual_dr.preprocessors[1], PlusOnePreprocessor)

    actual_train_data = actual_dr.train_generator
    actual_val_data = actual_dr.val_generator
    actual_test_data = actual_dr.test_generator

    expected_train_data = expected_dr.train_generator
    expected_val_data = expected_dr.val_generator
    expected_test_data = expected_dr.test_generator

    check_equal_data_generator(actual_train_data, expected_train_data)
    check_equal_data_generator(actual_val_data, expected_val_data)
    check_equal_data_generator(actual_test_data, expected_test_data)


def test_load_h5_data(h5file):
    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(self, x, y):
            return x + 1, y + 1

    actual_dr = load_data(load_json_config(
        read_file('tests/json/h5_dataset_config.json')))

    expected_dr = H5Reader(
        filename=h5file,
        batch_size=8,
        preprocessors=PlusOnePreprocessor(),
        x_name='input',
        y_name='target',
        train_folds=[0, 1, 2],
        val_folds=[3],
        test_folds=[4, 5]
    )

    assert isinstance(actual_dr, H5Reader)
    assert isinstance(actual_dr.preprocessors[0], PlusOnePreprocessor)

    actual_train_data = actual_dr.train_generator
    actual_val_data = actual_dr.val_generator
    actual_test_data = actual_dr.test_generator

    expected_train_data = expected_dr.train_generator
    expected_val_data = expected_dr.val_generator
    expected_test_data = expected_dr.test_generator

    check_equal_data_generator(actual_train_data, expected_train_data)
    check_equal_data_generator(actual_val_data, expected_val_data)
    check_equal_data_generator(actual_test_data, expected_test_data)


def test_load_h5_data_multi_preprocessor(h5file):
    @custom_preprocessor
    class PlusOnePreprocessor(BasePreprocessor):
        def transform(self, x, y):
            return x + 1, y + 1

    actual_dr = load_data(load_json_config(
        read_file('tests/json/h5_dataset_config_multi_preprocessor.json')))

    expected_dr = H5Reader(
        filename=h5file,
        batch_size=8,
        preprocessors=[PlusOnePreprocessor(), PlusOnePreprocessor()],
        x_name='input',
        y_name='target',
        train_folds=[0, 1, 2],
        val_folds=[3],
        test_folds=[4, 5]
    )

    assert isinstance(actual_dr, H5Reader)
    assert isinstance(actual_dr.preprocessors[0], PlusOnePreprocessor)
    assert isinstance(actual_dr.preprocessors[1], PlusOnePreprocessor)

    actual_train_data = actual_dr.train_generator
    actual_val_data = actual_dr.val_generator
    actual_test_data = actual_dr.test_generator

    expected_train_data = expected_dr.train_generator
    expected_val_data = expected_dr.val_generator
    expected_test_data = expected_dr.test_generator

    check_equal_data_generator(actual_train_data, expected_train_data)
    check_equal_data_generator(actual_val_data, expected_val_data)
    check_equal_data_generator(actual_test_data, expected_test_data)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"


import pytest
import numpy as np
from deoxys.keras.models import Model
from deoxys.keras.layers import Input, Dense, Flatten, Dropout, \
    BatchNormalization, Conv2D, Conv2DTranspose, MaxPooling2D, concatenate, \
    Conv3D, Conv3DTranspose, MaxPooling3D
from deoxys.loaders.architecture import BaseModelLoader, ModelLoaderFactory, \
    SequentialModelLoader, UnetModelLoader, DenseModelLoader, \
    Vnet, load_architecture
from deoxys.customize import register_architecture, custom_architecture
from deoxys.utils import read_file, load_json_config


@pytest.fixture(autouse=True)
def clear_singleton():
    ModelLoaderFactory._loaders = {
        'Sequential': SequentialModelLoader,
        'Unet': UnetModelLoader,
        'Vnet': Vnet,
        'Dense': DenseModelLoader
    }  # clear singleton


@pytest.fixture
def modelloader_class():
    class TestModelLoader(BaseModelLoader):
        pass

    yield TestModelLoader


def check_same_models(actual_model, expected_model):
    actual_config = actual_model.get_config()
    expected_config = expected_model.get_config()
    name_map = [(actual_model.name, expected_model.name)]
    for (actual_layer, expected_layer) in zip(
            actual_model.layers, expected_model.layers):
        name_map.append((actual_layer.name, expected_layer.name))

    actual_config_str = str(actual_config)

    for actual_name, expected_name in name_map:
        actual_config_str = actual_config_str.replace(
            "'" + actual_name + "'", "'" + expected_name + "'")

    return actual_config_str == str(expected_config)


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_architecture("TestModelLoader", object)


def test_register_modelloader_success(modelloader_class):
    register_architecture("TestModelLoader", modelloader_class)

    assert ModelLoaderFactory(
    )._loaders["TestModelLoader"] is modelloader_class


def test_register_duplicate_modelloader(modelloader_class):
    register_architecture("TestModelLoader", modelloader_class)

    with pytest.raises(KeyError):
        register_architecture("TestModelLoader", modelloader_class)


def test_decorator():
    @custom_architecture
    class TestModelLoader2(BaseModelLoader):
        pass

    assert ModelLoaderFactory._loaders["TestModelLoader2"] is TestModelLoader2


def test_load_sequential_model():
    architecture = load_json_config(
        read_file('tests/json/sequential_architecture.json'))

    input_params = {'shape': [32, 32]}

    model = load_architecture(architecture, input_params)

    input_layer = Input(shape=(32, 32))
    flatten = Flatten()(input_layer)
    dense = Dense(units=128, activation='relu')(flatten)
    dropout = Dropout(rate=0.2)(dense)
    output_layer = Dense(units=10, activation='softmax')(dropout)

    expected_model = Model(inputs=input_layer, outputs=output_layer)

    assert check_same_models(model, expected_model)


def test_load_unet_model():
    architecture = load_json_config(
        read_file('tests/json/unet_architecture.json')
    )

    input_params = {'shape': [128, 128, 3]}

    model = load_architecture(architecture, input_params)

    def conv_layers(filters, pre_layer):
        conv = BatchNormalization()(
            Conv2D(filters,
                   kernel_size=3,
                   activation='relu',
                   padding='same')(pre_layer))

        return BatchNormalization()(
            Conv2D(filters,
                   kernel_size=3,
                   activation='relu',
                   padding='same')(conv))

    input_layer = Input(shape=[128, 128, 3])

    conv_1 = conv_layers(4, input_layer)
    max_pool_1 = MaxPooling2D()(conv_1)

    conv_2 = conv_layers(8, max_pool_1)
    max_pool_2 = MaxPooling2D()(conv_2)

    conv_3 = conv_layers(16, max_pool_2)
    max_pool_3 = MaxPooling2D()(conv_3)

    conv_4 = conv_layers(32, max_pool_3)
    max_pool_4 = MaxPooling2D()(conv_4)

    conv_5 = conv_layers(64, max_pool_4)
    max_pool_5 = MaxPooling2D()(conv_5)

    conv_t_kwargs = {"kernel_size": 3,
                     "strides": [
                         2,
                         2
                     ],
                     "padding": "same"}

    conv_6 = conv_layers(128, max_pool_5)
    conv_trans_1 = Conv2DTranspose(32, **conv_t_kwargs)(conv_6)

    upconv_1 = conv_layers(64, concatenate([conv_5, conv_trans_1]))
    conv_trans_2 = Conv2DTranspose(16, **conv_t_kwargs)(upconv_1)

    upconv_2 = conv_layers(32, concatenate([conv_4, conv_trans_2]))
    conv_trans_3 = Conv2DTranspose(8, **conv_t_kwargs)(upconv_2)

    upconv_3 = conv_layers(16, concatenate([conv_3, conv_trans_3]))
    conv_trans_4 = Conv2DTranspose(4, **conv_t_kwargs)(upconv_3)

    upconv_4 = conv_layers(8, concatenate([conv_2, conv_trans_4]))
    conv_trans_5 = Conv2DTranspose(2, **conv_t_kwargs)(upconv_4)

    upconv_5 = conv_layers(4, concatenate([conv_1, conv_trans_5]))
    output = Conv2D(1, kernel_size=1, activation='sigmoid')(upconv_5)

    expected_model = Model(inputs=input_layer, outputs=output)

    assert check_same_models(model, expected_model)


def test_load_unet_model_resize():
    architecture = load_json_config(
        read_file('tests/json/unet_architecture.json')
    )

    input_params = {'shape': [129, 129, 3]}

    model = load_architecture(architecture, input_params)

    assert np.all(model.input_shape == (None, 129, 129, 3))
    assert np.all(model.output_shape == (None, 129, 129, 1))


def test_load_vnet_model():
    architecture = load_json_config(
        read_file('tests/json/vnet_architecture.json')
    )

    input_params = {'shape': [128, 128, 128, 3]}

    model = load_architecture(architecture, input_params)

    def conv_layers(filters, pre_layer):
        conv = BatchNormalization()(
            Conv3D(filters,
                   kernel_size=3,
                   activation='relu',
                   padding='same')(pre_layer))

        return BatchNormalization()(
            Conv3D(filters,
                   kernel_size=3,
                   activation='relu',
                   padding='same')(conv))

    input_layer = Input(shape=[128, 128, 128, 3])

    conv_1 = conv_layers(4, input_layer)
    max_pool_1 = MaxPooling3D()(conv_1)

    conv_2 = conv_layers(8, max_pool_1)
    max_pool_2 = MaxPooling3D()(conv_2)

    conv_3 = conv_layers(16, max_pool_2)
    max_pool_3 = MaxPooling3D()(conv_3)

    conv_4 = conv_layers(32, max_pool_3)
    max_pool_4 = MaxPooling3D()(conv_4)

    conv_5 = conv_layers(64, max_pool_4)
    max_pool_5 = MaxPooling3D()(conv_5)

    conv_t_kwargs = {"kernel_size": 3,
                     "strides": [
                         2,
                         2,
                         2
                     ],
                     "padding": "same"}

    conv_6 = conv_layers(128, max_pool_5)
    conv_trans_1 = Conv3DTranspose(32, **conv_t_kwargs)(conv_6)

    upconv_1 = conv_layers(64, concatenate([conv_5, conv_trans_1]))
    conv_trans_2 = Conv3DTranspose(16, **conv_t_kwargs)(upconv_1)

    upconv_2 = conv_layers(32, concatenate([conv_4, conv_trans_2]))
    conv_trans_3 = Conv3DTranspose(8, **conv_t_kwargs)(upconv_2)

    upconv_3 = conv_layers(16, concatenate([conv_3, conv_trans_3]))
    conv_trans_4 = Conv3DTranspose(4, **conv_t_kwargs)(upconv_3)

    upconv_4 = conv_layers(8, concatenate([conv_2, conv_trans_4]))
    conv_trans_5 = Conv3DTranspose(2, **conv_t_kwargs)(upconv_4)

    upconv_5 = conv_layers(4, concatenate([conv_1, conv_trans_5]))
    output = Conv3D(1, kernel_size=1, activation='sigmoid')(upconv_5)

    expected_model = Model(inputs=input_layer, outputs=output)

    assert check_same_models(model, expected_model)


def test_load_vnet_model_resize():
    architecture = load_json_config(
        read_file('tests/json/vnet_architecture.json')
    )

    input_params = {'shape': [129, 129, 129, 3]}

    model = load_architecture(architecture, input_params)

    assert np.all(model.input_shape == (None, 129, 129, 129, 3))
    assert np.all(model.output_shape == (None, 129, 129, 129, 1))
